unit Prefab;

interface
uses Classes,J_Level, geometry, lev_utils, FileOperations, Files,
     misc_utils, sysUtils;

Var
 Prefabs:TStringList;

Function GetPreFab(n:integer):TJKSector;
Procedure LoadPrefabs(const fname:string);

implementation

Function GetPreFab(n:integer):TJKSector;
begin
 Result:=TJKSector(Prefabs.Objects[n]);
end;

Procedure LoadPrefabs(const fname:string);
var p,i:integer;
    t:TTextFile;
    s,st,w:string;

Procedure ReadFmt(const fmt:string;v:array of pointer);
begin
 t.Readln(st);
 SScanf(st,fmt,v);
end;

Procedure ReadS(Var s:string);
begin
 t.Readln(s);
 RemoveComment(s);
end;

Function LoadSector:TJKSector;
var sec:TJKSector;
    sf,nsf,vx,nv,nsv:integer;
    asec,asf:integer;
    v:TJKVertex;
    surf:TJKSurface;
begin
 sec:=Level.NewSector;
 Result:=sec;
try
 With Sec do
 begin
  ReadFmt('%s %s %f %x %f %f %f %f %f %d',
   [@ColorMap,@Sound,@snd_vol,@Flags,@Ambient,@extra,@tint.dx,@tint.dy,@tint.dz,@Layer]);
  if ColorMap='-' then ColorMap:='';
  if Sound='-' then Sound:='';

  ReadFmt('VXS %d',[@nv]);
  for vx:=0 to nv-1 do
  begin
   V:=Sec.NewVertex;
   {Sec.Vertices.Add(v);}
   With V do ReadFmt('%f %f %f',[@x,@y,@z]);
  end;
  ReadFmt('SURFS %d',[@nsf]);
  for sf:=0 to nsf-1 do
  begin
   surf:=Sec.NewSurface;
   Sec.Surfaces.Add(Surf);
   With Surf do
   begin
    ReadFmt('%s %x %x %d %d %d %f %d %d %d %x',
     [@Material,@SurfFlags,@FaceFlags,@geo,@light,@tex,@extralight,@nsv,@asec,@asf,@AdjoinFlags]);
    if Material='-' then Material:='';
    {Num:=asec shl 16+(asf and $FFFF);}
    ReadS(st);
    p:=1;
    for vx:=0 to nsv-1 do
    begin
     p:=GetWord(st,p,w);
     ValInt(w,nv);
     AddVertex(Sec.Vertices[nv]);
     With TXVertices[vx] do
     begin
      p:=GetWord(st,p,w);
      ValSingle(w,u);
      p:=GetWord(st,p,w);
      ValSingle(w,v);
      p:=GetWord(st,p,w);
      ValSingle(w,Intensity);
     end;
    end;
    {Load RGB}
    for vx:=0 to nsv-1 do
    With TXVertices[vx] do
    begin
     p:=GetWord(st,p,w);
     if w='' then break;
     ValSingle(w,r);
     p:=GetWord(st,p,w);
     ValSingle(w,g);
     p:=GetWord(st,p,w);
     ValSingle(w,b);
    end;

   end;
  end;
 end;
 Sec.Renumber;
except
 On E:Exception do
 begin
  PanMessageFmt(mt_error,'Error loading %s at line %d: %s',
                [fname,t.curline,e.message]);
  sec.free; result:=nil;
 end;
end;
end;


var sec:TJKSector;
    sname:string;
begin
 for i:=1 to Prefabs.count-1 do Prefabs.Objects[i].Free;
 For i:=Prefabs.count-1 downto 1 do Prefabs.Delete(i);

 try
  t:=TTextFile.CreateRead(OpenFileRead(fname,0));
 except
 on E:Exception do
 begin
   PanMessage(mt_warning,E.Message+' '+fname);
   exit;
 end;
end;

 try

 while not t.eof do
 begin
  t.Readln(s);
  RemoveComment(s);
  p:=GetWord(s,1,w);
  if CompareText(w,'SEC')=0 then
  begin
   sname:=Trim(Copy(s,p,length(s)));
   sec:=LoadSector;
   if sec<>nil then Prefabs.AddObject(sname,sec);
  end;
 end;


 finally
  t.Fclose;
 end;

end;

Procedure InitList;
var js:TJKSector;
    z,x:TVector;
begin
 Prefabs:=TStringList.Create;
 js:=Level.NewSector;
 SetVec(z,0,0,1);
 SetVec(x,1,0,0);
 CreateCubeSec(js,0,0,0,z,x);
 Prefabs.AddObject('Cube',js);
end;

Initialization
InitList;

end.
